""" Render object """

# import logging
from typing import Any, Dict, List, Optional, Tuple, Union
from functools import lru_cache
# from enum import Enum, unique

from xml.etree.ElementTree import Element, SubElement
from xml.etree import ElementTree

from pydot import Dot, Node, Edge

from .storage import Storage
from .models import Pathway, Entry
from .resolver import (
    Resolver,
    # get_gene_names,
)
from .utils import ColorGradient



# TODO: add hex string to int tuple function
# TODO: add int tuple to hex string function


# Color variables
# TODO: implement all colors as tuple or Union[str, Tuple[int, int, int]]
# TODO: convert color from tuple to string on function level

# @unique
# class Color(Enum):
#     """
#     Color variables.
#     """
#     WHITE: Tuple[int, int, int] = (255, 255, 255)
#     BLACK: Tuple[int, int, int] = (0, 0, 0)


# TODO: use enum nested classes for different parameter options in pydot ???
# class PydotNode(Enum):
#     """
#     ...
#     """
#     class LineStyle(Enum):
#         """
#         ...
#         """
#         DASHED: str = ""



# Helper functions for renderer

def generate_embedded_html_table(
    items: Dict[str, str],
    border: int = 0,
    cellborder: int = 1,
    truncate: Optional[int] = None,
    ) -> str:
    """
    Generate HTML table in insert into label of dot node.

    `generate_embedded_html_table({"gene1": "#ffffff", "gene2": "#454545"})`

    :param typing.Dict[str, str] items: Items are dicts with have format `{name: hex_color}`.
    :param int border: Thickness of table border. (Default: 0)
    :param int cellborder: Thickness of cell border within the table. (Default: 1)
    :param typing.Optional[int] truncate: Maximal number of items in table. Set to None to disable trunaction. \
        (Default: None)
    :return: Returns html string of table.
    :rtype: str
    """

    # TODO: items input must be an ordered iterable!

    # TODO: implement more suppored html attributes in table, tr and td elements

    element_table: Element = Element(
        "table",
        attrib={
            "border": str(border),
            "cellborder": str(cellborder),
            "cellspacing": "0",
            "cellpadding": "4",
        },
    )


    # TODO: implement multiple cols for longer lists (square format)
    for index, (key, value) in enumerate(items.items()):

        element_row: Element = SubElement(element_table, "tr", attrib={})

        element_col: Element = SubElement(element_row, "td")

        if truncate is not None and index >= truncate:

            element_col.text = f"{len(items.items()) - truncate} more genes..."
            break

        element_col.attrib = {"bgcolor": value}

        # Set key as inner text of table cell
        element_col.text = key

    return ElementTree.tostring(element_table).decode("utf-8")




class Renderer:
    """
    Renderer for KEGG Pathway.
    """

    def __init__(
        self,
        kegg_pathway: Pathway,
        gene_dict: Optional[Dict[str, float]] = None,
        cache_or_resolver: Optional[Union[Storage, str, Resolver]] = None,
        # resolve_compounds: bool = True # TODO: Specify if renderer should resolver compounds in human readable text

        upper_color: Tuple[int, int, int] = (255, 0, 0),
        lower_color: Tuple[int, int, int] = (0, 0, 255),
    ) -> None:
        """
        Init Renderer instance for KEGG Pathway.

        :param Pathway kegg_pathway: Pathway instance to render.
        :param typing.Optional[typing.Dict[str, float]] gene_dict: Dict to specify overlay color \
            gradient to rendered entries.
        :param typing.Optional[typing.Union[Storage, str, cache_or_resolver]] cache: \
            Specify cache for resolver instance or pass resolver. Resolver is needed to get compound data needed for \
            rendering.
        :param typing.Tuple[int, int, int] upper_color: Color for upper bound of color gradient.
        :param typing.Tuple[int, int, int] lower_color: Color for lower bound of color gradient.
        """

        # Pathway instance to render
        self.pathway: Pathway = kegg_pathway

        # Generate pydot Graph instance
        self.graph: Dot = Dot(
            'pathway',
            graph_type='digraph',
            bgcolor='#ffffff',
            labelloc="t",
            label=self.pathway.title,
            fontsize=25,
            rankdir="TB",
            splines="normal", # "ortho"/line
            arrowhead="normal",
        )

        # overlay vars
        self.overlay: Dict[str, float] = {}

        if gene_dict is not None:
            self.overlay = gene_dict


        # TODO: move to render function ??
        # Generate color map
        self.upper_color: tuple = upper_color
        self.lower_color: tuple = lower_color


        # Init resolver instance from pathway org code.
        resolver_buffer: Optional[Resolver] = None

        if isinstance(cache_or_resolver, (str, Storage)) or cache_or_resolver is None:
            resolver_buffer = Resolver(cache=cache_or_resolver)
        elif isinstance(cache_or_resolver, Resolver):
            resolver_buffer = cache_or_resolver
        else:
            # Raise error of type is not correct
            raise TypeError("String to directory, storage instance or resolver instance must be passed.")


        self.resolver: Resolver = resolver_buffer


    # implement color gradient as properties with lru_cache decorator
    @property
    def cmap_upreg(self) -> List[str]:
        """
        Generated color map as list of hexadecimal strings for upregulated genes in gene dict.
        """
        @lru_cache(maxsize=1)
        def cache_wrapper() -> List[str]:
            return ColorGradient(
                start=(255, 255, 255),
                stop=self.upper_color,
                steps=100
            ).get_list()

        return cache_wrapper()


    @property
    def cmap_downreg(self) -> List[str]:
        """
        Generated color map as list of hexadecimal strings for downregulated genes in gene dict.
        """
        @lru_cache(maxsize=1)
        def cache_wrapper() -> List[str]:
            return ColorGradient(
                start=(255, 255, 255),
                stop=self.lower_color,
                steps=100
            ).get_list()

        return cache_wrapper()


    def get_gene_color(self, gene_id: str, default_color: Tuple[int, int, int] = (255, 255, 255)) -> str:
        """
        Get overlay color for given gene.

        :param str gene_id: Identify of gene.
        :param typing.Tuple[int, int, int] default_color: Default color to return if gene is not found in gene_dict. \
            Format in RGB tuple.
        :return: Color of gene by expression level specified in gene_dict.
        :rtype: str
        """

        # Return default color if gene is not found
        if self.overlay.get(gene_id) in (None, 0.0):
            return ColorGradient.to_hex(color=default_color)



        # Get expression limits
        exp_min: float = min(self.overlay.values())
        exp_max: float = max(self.overlay.values())

        if self.overlay[gene_id] < 0:
            # Expression below 0 (Downregulation)
            return self.cmap_downreg[abs(int(self.overlay[gene_id] / exp_min * 100))]

        # Expression above 0 (Upregulation)
        return self.cmap_upreg[abs(int(self.overlay[gene_id] / exp_max * 100))]


    # TODO: implement not here
    # def resolve_missing_gene_names(
    #     self,
    #     truncate_gene_list: Optional[int] = None,
    #     ) -> Dict[str, str]:
    #     """
    #     Resolve names of gene entries with only gene id given, but no human-readable names.
    #     This cases appears in entries with type gene with multiple, space-seperated gene id entries in the name
    #     attribute.

    #     :param typing.Optional[int] truncate_gene_list: With truncate entries with multiple space-seperated names \
    #         to given length. To keep all genes, set parameter to `None`.
    #     """

    #     # TODO: get list of genes with only kegg id and no label
    #     gene_names: List[str] = []

    #     # get all gene ids from entries with multiples names
    #     for entry in self.pathway.entries:
    #         if entry.type == "gene":
    #             name_parts: List[str] = entry.name.split(" ")
    #             if len(name_parts) > 1:

    #                 max_name_index: int = len(name_parts)

    #                 # Overwrite max index of list iteration, if truncation is set to integer below list length
    #                 if truncate_gene_list is not None and truncate_gene_list < max_name_index:
    #                     max_name_index = truncate_gene_list

    #                 for name_index in range(1, max_name_index):
    #                     if name_parts[name_index] not in gene_names:
    #                         gene_names.append(name_parts[name_index])

    #     # save in instance list of gene-id to gene name lookup dict
    #     # TODO: check if all genes got resolved

    #     loopup_dict: Dict[str, str] = {}

    #     # Split organism code from gene id
    #     for key, value in get_gene_names(genes=gene_names, max_genes=len(gene_names) + 1).items():
    #         loopup_dict[key.split(":")[1]] = value

    #     return loopup_dict


    def render(
        self,
        # resolve_unlabeled_genes: bool = True,
        display_unlabeled_genes: bool = True,
        # truncate_gene_list: Optional[int] = None,
        ) -> None:
        """
        Render KEGG pathway.


        :param bool display_unlabeled_genes: Entries in the KGML format can have space-seperated entry names. \
            Set this parameter to `False` to hide the entries.
        """

        # :param bool resolve_unlabeled_genes: If `True` the function will resolve all gene names for gene entries \
        #     which only have a gene id given.
        # :param typing.Optional[int] truncate_gene_list: With truncate entries with multiple space-seperated names \
        #     to given length. To keep all genes, set parameter to `None`.
        # """

        # pylint: disable=too-many-branches,too-many-locals,too-many-statements

        # resolved_gene_names: Dict[str, str] = {}
        # if resolve_unlabeled_genes is True:
        #     resolved_gene_names = self.resolve_missing_gene_names(truncate_gene_list=truncate_gene_list)

        # add all nodes and edges
        related_entries = [int(p.entry1) for p in self.pathway.relations]
        related_entries.extend([int(p.entry2) for p in self.pathway.relations])

        for entry in self.pathway.entries:

            # Use entry id as default label
            # TODO: use different default label as fallback
            entry_label: str = entry.id

            # only render genes with at least 1 relation
            if int(entry.id) in related_entries:
                # case select for types gene, comp, group, ...
                if entry.type == "gene":


                    # Get entry name by graphics element name attribute
                    if entry.graphics is not None and entry.graphics.name is not None:
                        entry_label = entry.graphics.name.split(", ")[0]


                    # Check if entry name contains multiple gene entries
                    # TODO: request the names of the entry names with .../find/gene1+gene2 ->
                    # parse list and use names as labels

                    if len(entry.get_gene_id()) > 1 and display_unlabeled_genes is True:
                        entry_gene_list: List[str] = entry.get_gene_id()

                        # Overwrite name of first item with graphics name
                        entry_gene_list[0] = entry_label

                        # TODO: resolve missing gene names by lookup dict
                        # Skip first entry
                        # for index in range(1, len(entry_gene_list)):
                        #     # Get item from lookup dict and fallback to gene id
                        #     # TODO: mark entry if gene id is used ??
                        #     entry_gene_list[index] = resolved_gene_names.get(
                        #         entry_gene_list[index], entry_gene_list[index]
                        #     )


                        # Add node
                        self.graph.add_node(Node(
                            name=entry.id,
                            label="<" + generate_embedded_html_table(
                                items={ gene_id: self.get_gene_color(gene_id) for gene_id in entry_gene_list }
                            ) + ">",
                            shape="plaintext",
                            style="filled",
                            color="#000000",
                            fillcolor="#ffffff",
                        ))


                    else:

                        # Single node
                        self.graph.add_node(Node(
                            name=entry.id,
                            label=entry_label,
                            shape="rectangle",
                            style="filled",
                            color="#000000",
                            fillcolor=self.get_gene_color(gene_id=entry.get_gene_id()[0]),
                        ))



                elif entry.type == "group":

                    component_label: List[Tuple[str, str]] = []

                    # Iterate of components of group entry
                    for comp in entry.components:

                        component_entry: Optional[Entry] = self.pathway.get_entry_by_id(comp.id)

                        if component_entry is not None and \
                            component_entry.graphics is not None and \
                            component_entry.graphics.name is not None:

                            # Append tuple of gene id and color to labels
                            component_label.append((
                                    component_entry.graphics.name.split(", ")[0],
                                    self.get_gene_color(component_entry.get_gene_id()[0]),
                            ))


                    # Generate html table string from gene dict
                    # TODO: adjust cell spacing/padding, border

                    html_table_string: str = generate_embedded_html_table(
                        items=dict(component_label),
                    )

                    # Add dot node to graph
                    self.graph.add_node(Node(
                        name=entry.id,
                        label=f"<{html_table_string}>",
                        # shape="rectangle",
                        shape="plaintext",
                        style="filled",
                        color="#000000",
                        fillcolor="#ffffff",
                    ))


                elif entry.type == "compound":

                    # TODO: resolver compound with resolver


                    if entry.graphics is not None and entry.graphics.name is not None:
                        entry_label = entry.graphics.name.split(", ")[0]

                    # Add compound node to graph
                    self.graph.add_node(Node(
                        name=entry.id,
                        label=entry_label,
                        shape="oval",
                        style="filled",
                        color="#000000",
                        fillcolor="#ffffff",
                    ))


                elif entry.type == "reaction":

                    # TODO: implement reaction rendering

                    pass


                # TODO: add pathway map to dot graph
                # elif entry.type == "map":

                #     if entry.graphics is not None and entry.graphics.name is not None:
                #         entry_label = entry.graphics.name # .split(", ")[0]

                #     # Add map node to graph
                #     self.graph.add_node(Node(
                #         name=entry.id,
                #         label=entry_label,
                #         shape="rectangle",
                #         style="rounded,filled",
                #         color="#000000",
                #         fillcolor="#ffffff",
                #     ))



        for rel in self.pathway.relations:

            # Create edge instance from relation enties
            relation_edge: Edge = Edge(
                src=rel.entry1,
                dst=rel.entry2,
            )

            # default line style
            line_style: str = "solid"


            # Check for type of interaction and set line style of edge
            if rel.type == "GErel":
                # Gene expression interaction gets a distinct line style
                line_style = "dashed"

            elif rel.type == "PCrel":
                # Protein-compound interaction gets a distinct line style
                line_style = "dotted"



            # Add molecular events as edge labels
            molecular_event_dict: Dict[str, str] = {
                "phosphorylation": "+p",
                "dephosphorylation": "-p",
                "glycosylation": "+g",
                "ubiquitination": "+u",
                "methylation": "+m",
            }

            # Iterate over relation subtypes and check if moleuclar event is present
            edge_label: Optional[str] = None

            # Default arrowhead
            arrowhead: str = "none"

            for subtype in rel.subtypes:

                # Check if subtype is molecular event to set label of edge
                if subtype.name in molecular_event_dict:

                    # Check if label is none (no subtype found in prior iter steps)
                    if edge_label is None:
                        edge_label = molecular_event_dict[subtype.name]
                    else:
                        # TODO: is this event even possible ?
                        # TODO: handle but call warning
                        # A molecular event was already found. Append label to existing label
                        edge_label += ", " + molecular_event_dict[subtype.name]


                # Check for quality of interaction to set arrow head of edge
                if subtype.name in ("activation", "expression", "indirect effect", "binding/association"):
                    arrowhead = "normal"

                elif subtype.name in ("inhibition", "repression", "dissociation"):
                    arrowhead = "tee"

                elif subtype.name == "state change":
                    arrowhead = "diamond"


            # set arrowhead to edge instance
            relation_edge.set(name="arrowhead", value=arrowhead)


            # Set label if molecular event is found in relation subtypes
            if edge_label is not None:
                relation_edge.set(name="label", value=edge_label)

            # Set line style to edge instance
            relation_edge.set(name="style", value=line_style)

            # Add edge to graph
            self.graph.add_edge(relation_edge)




    def to_string(self) -> str:
        """
        pydot graph instance to dot string.

        :return: Generated dot string of pathway.
        :rtype: str
        """

        # Generate dot string from pydot graph object
        render_string: Any = self.graph.to_string()


        # Check correct type of dot string
        if not isinstance(render_string, str):
            raise TypeError("Object returned from pydot graph object is not a string.")

        return render_string



    def to_binary(self, extension: str) -> bytes:
        """
        Export pydot graph to binary data.

        :param str extension: Extension of file to export. Use format string like "png", "svg", "pdf" or "jpeg".
        :return: File content are bytes object.
        :rtype: bytes
        :raises TypeError: If variable with generated dot graph is not type bytes.
        """

        # render with pydot to binary
        graph_data: Any = self.graph.create(prog="dot", format=extension)

        # Type check of return value
        if not isinstance(graph_data, bytes):
            raise TypeError("Failed to create binary file object from pydot graph instance.")

        return graph_data


    def to_file(self, filename: str, extension: str) -> None:
        """
        Export pydot graph to file.

        :param str filename: Filename to save file at.
        :param str extension: Extension of file to export. Use format string like "png", "svg", "pdf" or "jpeg".
        """

        # TODO: get export format from filename

        # Get binary data from graph object
        binary_data: bytes = self.to_binary(extension=extension)

        # Save binary data to file object
        with open(filename, mode="wb") as file_obj:
            file_obj.write(binary_data)



""" KEGG Enrichment analysis core """

from csv import DictWriter
import os
from typing import Dict, List, Any, Union, Optional
from io import IOBase


from scipy import stats

from .models import Pathway


class EnrichmentResult:
    """
    Results of KEGG pathway enrichment analysis.
    """

    # pylint: disable=too-many-instance-attributes,too-many-arguments
    def __init__(
        self,
        org: str,
        pathway_id: str,
        pathway_name: str,
        found_genes: list,
        pathway_genes: list,
        pathway_title: Optional[str] = None,
    ) -> None:

        """
        Init Result of KEGG pathway enrichment analysis.

        :param str org: 3 letter code of organism used by KEGG database.
        :param str pathway_id: Identifier of KEGG pathway.
        :param str pathway_name: Name of KEGG pathway.
        :param list found_genes: List of found genes.
        :param list pathway_genes: List of all genes in pathway.
        """

        # Pathway descriptions
        self.organism: str = org
        self.pathway_id: str = pathway_id
        self.pathway_name: str = pathway_name
        self.pathway_title: Optional[str] = pathway_title

        # Results from enrichment analysis
        self.found_genes: list = found_genes
        self.pathway_genes: list = pathway_genes
        self.pvalue: Optional[float] = None


    @property
    def pathway_genes_count(self) -> int:
        """
        Count of pathway genes.

        :rtype: int
        :return: Number of genes in pathway.
        """
        return len(self.pathway_genes)


    @property
    def study_count(self) -> int:
        """
        Count of study genes.

        :rtype: int
        :return: Number of genes found in analysis of pathway.
        """
        return len(self.found_genes)


    def __str__(self) -> str:
        """
        Build string summary of KEGG path analysis result instance.
        :rtype: str
        :return: Returns string that describes the enrichment result instance.
        """
        return f"<EnrichmentResult {self.organism}:{self.pathway_id}" \
                f" ({self.pathway_name}) {len(self.found_genes)}/{self.pathway_genes_count}>"



    def json_summary(self, gene_delimiter: str = ",") -> Dict[str, Any]:
        """
        Build json summary for enrichment analysis.

        :param str gene_delimiter: Delimiter to seperate genes in gene list.
        :rtype: typing.Dict[str, typing.Any]
        :return: Summary of enrichment result instance as dict.
        """

        return {
            "pathway_name": self.pathway_name,
            "pathway_title": self.pathway_title,
            "pathway_id": self.pathway_id,
            "study_count": self.study_count,
            "pathway_genes": self.pathway_genes_count,
            "pvalue": self.pvalue,
            "found_genes": gene_delimiter.join([str(a) for a in self.found_genes])
        }


    @staticmethod
    def get_header() -> List[str]:
        """
        Build default header for enrichment analysis.

        :rtype: typing.List[str]
        :return: List of header names as string.
        """

        return [
            "pathway_name",
            "pathway_title",
            "pathway_id",
            "study_count",
            "pathway_genes",
            "pvalue",
            "found_genes",
        ]




class Enrichment:
    """
    KEGG pathway enrichment analysis.
    """

    def __init__(
        self,
        # org: str,
        pathways: List[Pathway],
    ) -> None:
        """
        Init KEGG pathway enrichment analysis.

        :param str org: Organism identifier used by KEGG database \
            (3 letter code, e.g. "mmu" for mus musculus or "hsa" for human).
        :param typing.List[Pathway] pathways: (Optional) List of Pathway instances or list of KEGG pathway identifier.
        """

        # self.organism: str = org
        # self.resolver: Resolver = Resolver(self.organism)

        self.result: List[EnrichmentResult] = []

        # Create pathway list
        self.all_pathways: List[Pathway] = pathways



    def _check_analysis_result_exist(self) -> None:
        """
        Check if summary exists.
        """

        if not self.result or len(self.result) == 0:
            raise ValueError("need to 'run_analysis' first")


    def get_subset(self, subset: List[str], inplace: bool = False) -> List[EnrichmentResult]:
        """
        Create subset of analysis result by list of pathway ids

        :param typing.List[str] subset: List of pathway identifer to filter enrichment result by.
        :param bool inplace: Update instance variable of enrichment result list and overwrite with generated subset.
        :return: Subset of enrichment results.
        :rtype: typing.List[EnrichmentResult]
        """
        self._check_analysis_result_exist()

        buffer = []
        # subset = [str(s) for s in subset]
        for item in self.result:
            if str(item.pathway_id) in subset:
                buffer.append(item)

        if inplace is True:
            self.result = buffer

        return buffer


    def run_analysis(self, gene_list: List[str]) -> List[EnrichmentResult]:
        """
        List of gene ids. Return list of EnrichmentResult instances

        :param typing.List[str] gene_list: List of genes to analyse.
        :return: List of enrichment result instances.
        :rtype: typing.List[EnrichmentResult]
        """
        # pylint: disable=too-many-locals

        all_found_genes: int = 0
        absolute_pathway_genes: int = 0
        study_n: int = len(gene_list)

        for pathway in self.all_pathways:

            genes_found: List[str] = []
            all_pathways_genes = pathway.get_genes()
            absolute_pathway_genes += len(all_pathways_genes)


            # Check for intersection between gene list and genes in pathway
            for gene_id in all_pathways_genes:
                if gene_id in gene_list:
                    genes_found.append(gene_id)

            all_found_genes += len(genes_found)

            # Create analysis results instance and append to list of results
            pathway_result: EnrichmentResult = EnrichmentResult(
                org=pathway.org,
                pathway_id=pathway.number,
                pathway_name=pathway.name,
                pathway_title=pathway.title,
                found_genes=genes_found,
                pathway_genes=all_pathways_genes
            )
            self.result.append(pathway_result)

        # Perform Fisher exact test
        # http://docs.scipy.org/doc/scipy-0.17.0/reference/generated/scipy.stats.fisher_exact.html
        for analysis in self.result:

            # Skip p value calculation if no genes are found
            if analysis.study_count > 0:

                a_var: int = analysis.study_count
                b_var: int = study_n - analysis.study_count
                c_var: int = analysis.pathway_genes_count - analysis.study_count
                d_var: int = absolute_pathway_genes - analysis.pathway_genes_count - b_var


                # Calculate p value with Fisher exact test
                _, pval = stats.fisher_exact(
                    [
                        [a_var, b_var],
                        [c_var, d_var],
                    ]
                )

                analysis.pvalue = pval


        return self.result


    def to_json(self) -> List[Dict[str, Any]]:
        """
        Export to json dict.

        :rtype: typing.List[typing.Dict[str, typing.Any]]
        :return: Json dict of enrichment results.
        """

        self._check_analysis_result_exist()

        result: list = []

        for item in self.result:
            result.append(item.json_summary())

        return result


    def to_csv(
        self,
        file_obj: Union[str, IOBase, Any],
        delimiter: str = "\t",
        overwrite: bool = False
        ) -> None:
        """
        Save result summary as file.

        :param typing.Union[str, io.IOBase, typing.Any] file_obj: String to file or IOBase object
        :param str delimiter: Deleimiter used for csv.
        :param bool overwrite: Set to True to overwrite file, if already exist.
        """

        # Check if summary exists
        self._check_analysis_result_exist()

        csv_file: Optional[IOBase] = None

        if isinstance(file_obj, str):
            # file is str (Name of file)
            if os.path.isfile(file_obj) and not overwrite:
                raise RuntimeError(
                    f"File {file_obj} does already exist." \
                    "To solve please set overwrite=True."
                )

            # pylint: disable=consider-using-with
            csv_file = open(file_obj, mode="w", encoding="utf-8")
        elif isinstance(file_obj, IOBase):
            # file is IOBase (File object stream)
            csv_file = file_obj

        else:
            raise TypeError("Argument 'file_obj' must be string or IOBase instance, like an open file object.")



        # Delimiter of gene names
        child_delimiter = " "

        # Change child delimiter to aviod csv conficts
        if child_delimiter == delimiter:
            raise ValueError("This delimiter is reserved to seperate list of genes.")

        headers: List[str] = EnrichmentResult.get_header()
        writer: DictWriter = DictWriter(csv_file, fieldnames=headers, delimiter=delimiter)

        for item in self.result:
            # Write lines
            writer.writerow(item.json_summary(gene_delimiter=child_delimiter))

        csv_file.close()





    def to_dataframe(self) -> Any:
        """
        Return analysis result as pandas DataFrame. Required pandas dependency.

        :return: Export enrichment results as pandas dataframe.
        :rtype: pandas.DataFrame
        """

        # Ignore import lint at this place to keep pandas an optional dependency
        # pylint: disable=import-outside-toplevel
        import pandas

        # Check if summary exists
        self._check_analysis_result_exist()
        summary_list = []
        for result in self.result:
            summary_list.append(result.json_summary())
        return pandas.DataFrame(summary_list)



""" Resolve requests to KEGG data Api """

from typing import Dict, List, Optional, Union
# from warnings import warn

import requests

from .utils import (
    parse_tsv_to_dict,
    # is_valid_gene_name,
)
from .storage import Storage
from .models import Pathway


def _request(url: str) -> str:
    """
    Url request helper function.

    :param str url: Url to request from.
    """

    response = requests.get(url=url)
    response.raise_for_status()
    return response.content.decode(encoding="utf-8")



def get_gene_names(genes: List[str], max_genes: int = 50) -> Dict[str, str]:
    """
    Resolve KEGG gene identifer to name using to KEGG database REST Api.
    Function is implemented outside the resolver instance, because requests are not cached and only gene identifier
    are used.

    :param typing.List[str] genes: List of gene identifer in format "<organism>:<code>"
    :return: Dict of gene idenifier to gene name.
    :rtype: typing.Dict[str, str]
    """

    if len(genes) == 0:
        raise ValueError("No items to request.")

    # Check maximum number of entries requests
    # TODO: build fallback to make multiple requests from long lists (iterate over list chunks)
    if len(genes) > max_genes:
        raise ValueError(f"Too many entries are requested at once ({len(genes)}/50).")

    # TODO check if pattern of identifer is correct
    # for item in genes:
    #     if not is_valid_gene_name(value=item):
    #         raise ValueError(
    #             f"Item '{item}' is not a valid gene identifer." \
    #             "Identifier must be 3 letter organism code with 5 digit KEGG gene id."
    #         )

    # Build query string
    query_string: str = "+".join(genes)

    # Request without cache
    resolve_dict: Dict[str, str] = parse_tsv_to_dict(data=_request(f"http://rest.kegg.jp/list/{query_string}"))

    # Sanitize dict by splitting first entry of gene name
    result_dict: Dict[str, str] = {}

    for key, value in resolve_dict.items():
        result_dict[key] = value.split(", ")[0]


    # Check if all genes are in dict
    # for item in genes:
    #     if item not in result_dict:
    #         warn(
    #             message=f"Gene identifer '{item}' could not be resolved by API request.",
    #             category=UserWarning,
    #         )


    return result_dict


class Resolver:
    """
    KEGG pathway resolver class.
    Request interface for KEGG API endpoint.
    """

    def __init__(
        self,
        cache: Optional[Union[Storage, str]] = None
    ) -> None:
        """
        Init Resolver instance.

        :param typing.Optional[typing.Union[Storage, str]] cache: Directory to use as cache storage or Storage instance.
        """

        # Handle different types of argument for cache

        _store: Optional[Storage] = None

        if isinstance(cache, str):
            _store = Storage(cachedir=cache)

        elif isinstance(cache, Storage):
            _store = cache
        else:
            # Fallback to default storage with hard coded folder name
            _store = Storage()


        # Internal storage instance
        self.storage: Storage = _store


    def _cache_or_request(
        self,
        filename: str,
        url: str,
        ) -> str:
        """
        Load file from cache folder. If file does not exist, request from given url.

        :param str filename: Filename to store in cache folder.
        :param str url: Url to online resource to request if file is not present in cache folder.
        :return: Returns content of file as string.
        :rtype: str
        """

        file_data: Optional[str] = None

        if self.storage.exist(filename=filename):

            # return pathway list dump
            file_data = self.storage.load(filename=filename)

        else:

            # Data not found in cache. Request from REST api

            file_data = _request(url=url)

            # Save in storage
            self.storage.save(filename=filename, data=file_data)

        return file_data




    def get_pathway_list(
        self,
        organism: str,
        ) -> Dict[str, str]:
        """
        Request list of pathways linked to organism.

        :param str organism: 3 letter organism code used by KEGG database.
        :return: Dict in format {<pathway-id>: <name>}.
        :rtype: typing.Dict[str, str]
        """

        # TODO: return as list of pathway identifier ?

        # TODO: verify org code

        # path:mmu00010	Glycolysis / Gluconeogenesis - Mus musculus (mouse)
        # path:<org><code>\t<name> - <org>


        list_data: str = self._cache_or_request(
            filename=f"pathway_list_{organism}.tsv",
            url=f"http://rest.kegg.jp/list/pathway/{organism}"
        )


        pathways: Dict[str, str] = parse_tsv_to_dict(
            data=list_data,
        )

        # return pathway list
        return pathways


    def get_pathway(
        self,
        organism: str,
        code: str,
        ) -> Pathway:
        """
        Load and parse KGML pathway by identifier.

        :param str organism: 3 letter organism code used by KEGG database.
        :param str code: Pathway identify used by KEGG database.
        :return: Returns parsed Pathway instance.
        :rtype: Pathway
        """

        # TODO: verify org code

        data: str = self._cache_or_request(
            filename=f"{organism}_path{code}.kgml",
            url=f"http://rest.kegg.jp/get/{organism}{code}/kgml",
        )

        # Parse string
        return Pathway.parse(data)



    def get_compounds(self) -> Dict[str, str]:
        """
        Get dict of components. Request from KEGG API if not in cache.

        :return: Dict of compound identifier to compound name.
        :rtype: typing.Dict[str, str]
        """

        compound_data: str = self._cache_or_request(
            filename="compound.tsv",
            url="http://rest.kegg.jp/list/compound",
        )

        # Parse tsv string
        result: Dict[str, str] = parse_tsv_to_dict(data=compound_data)

        return result



    def get_organism_list(self) -> Dict[str, str]:
        """
        Get organism codes from file or KEGG API.

        :return: Dict with format {<org>: <org-name>}
        :rtype: typing.Dict[str, str]
        """

        data: str = self._cache_or_request(
            filename="organism.tsv",
            url="http://rest.kegg.jp/list/organism"
        )

        result: Dict[str, str] = parse_tsv_to_dict(
            data=data,
            col_keys=1,
            col_values=2,
        )

        return result


    def check_organism(self, organism: str) -> bool:
        """
        Check if organism code exist.

        :param str organism: 3 letter organism code used by KEGG database.
        :return: Returns True if organism code is found in list of valid organisms.
        :rtype: bool
        """

        organism_list = self.get_organism_list()
        return organism_list.get(organism) is not None



""" Global constants """


# Example list for immune system pathways
# TODO: complete list of all pathways by groups

from typing import Dict, List


IMMUNE_SYSTEM_PATHWAYS: Dict[str, str] = {
    "04640": "Hematopoietic cell lineage",
    "04610": "Complement and coagulation cascades",
    "04611": "Platelet activation",
    "04620": "Toll-like receptor signaling pathway",
    "04621": "NOD-like receptor signaling pathway",
    "04622": "RIG-I-like receptor signaling pathway",
    "04623": "Cytosolic DNA-sensing pathway",
    "04625": "C-type lectin receptor signaling pathway",
    "04650": "Natural killer cell mediated cytotoxicity",
    "04612": "Antigen processing and presentation",
    "04660": "T cell receptor signaling pathway",
    "04658": "Th1 and Th2 cell differentiation",
    "04659": "Th17 cell differentiation",
    "04657": "IL-17 signaling pathway",
    "04662": "B cell receptor signaling pathway",
    "04664": "Fc epsilon RI signaling pathway",
    "04666": "Fc gamma R-mediated phagocytosis",
    "04670": "Leukocyte transendothelial migration",
    "04672": "Intestinal immune network for IgA production",
    "04062": "Chemokine signaling pathway"
}


# Constant element types in the KGML format
# Details at KGML manual https://www.kegg.jp/kegg/xml/docs/


RELATION_SUBTYPES: List[str] = [
    "compound",
    "hidden compound",
    "activation",
    "inhibition",
    "expression",
    "repression",
    "indirect effect",
    "state change",
    "binding/association",
    "dissociation",
    "missing interaction",
    "phosphorylation",
    "dephosphorylation",
    "glycosylation",
    "ubiquitination",
    "methylation",
]


RELATION_TYPES: List[str] = [
    "ECrel",
    "PPrel",
    "GErel",
    "PCrel",
    "maplink",
]

GRAPHIC_TYPE: List[str] = [
    "rectangle",
    "circle",
    "roundrectangle",
    "line"
]

ENTRY_TYPE: List[str] = [
    "ortholog",
    "enzyme",
    "reaction",
    "gene",
    "group",
    "compound",
    "map",
    "brite",
    "other"
]


REACTION_TYPE: List[str] = [
    "reversible",
    "irreversible",
]



""" Init keggtools module """


__version__: str = "1.0.0"

from .analysis import EnrichmentResult, Enrichment
from .const import IMMUNE_SYSTEM_PATHWAYS
from .models import Pathway, Relation, Entry, Graphics, Subtype, Component
from .render import Renderer
from .resolver import Resolver
from .storage import Storage
from .utils import ColorGradient



""" Storage of KEGG data. Caching downloaded files from API to local file system. """

import os
import pickle
from typing import Any, Optional


class Storage:
    """
    Storage handler class.
    """

    def __init__(self, cachedir: Optional[str] = None) -> None:
        """
        Init KEGG data storage instance.

        :param typing.Optional[str] cachedir: Path to folder to use as cache.
        """

        if cachedir is None:
            # Cachedir argument not given. Fallback to default cache directory
            cachedir = os.path.join(os.getcwd(), ".keggtools_cache")


        if os.path.isdir(cachedir) is False:
            # Directory does not exist. Auto-generate diretory
            os.mkdir(cachedir)
            # logging.info("Cache folder '%s' does not exist. Auto-generating folder.", cachedir)
            # raise NotADirectoryError(f"Directory '{cachedir}' does not exist.")


        self.cachedir = cachedir



    def check_cache_dir(self) -> None:
        """
        Checks if cache dir exist. Raises "NotADirectoryError" of caching folder not found.

        :raises NotADirectoryError: Error if cache folder does not exist.
        """

        if os.path.isdir(self.cachedir) is False:
            raise NotADirectoryError(f"Directory '{self.cachedir}' does not exist.")



    def build_cache_path(self, filename: str) -> str:
        """
        Build absolute filename for caching directory.

        :param str filename: Name of file (is used as suffix to cache directory).
        :return: Full filename with is inside cache folder.
        :rtype: str
        """

        self.check_cache_dir()
        return os.path.join(self.cachedir, filename)


    def exist(self, filename: str) -> bool:
        """
        Check if filename exist in caching dir.

        :param str filename: Filename to check.
        :return: Returns True if file with given name exist in cachedir.
        :rtype: bool
        """

        self.check_cache_dir()
        return os.path.isfile(os.path.join(self.cachedir, filename))


    def save(self, filename: str, data: str) -> str:
        """
        Save string as file in local storage. Returns absolute filename of save file.

        :param str filename: Filename to storage file at.
        :param str data: String data to save to cache file.
        :return: Full filename to cached file.
        :rtype: str
        """

        self.check_cache_dir()
        path: str = self.build_cache_path(filename=filename)

        with open(path, "w", encoding="utf-8") as f_obj:
            f_obj.write(data)
            f_obj.close()

        return path


    def save_dump(self, filename: str, data: Any) -> str:
        """
        Save binary dump as file in local storage. Returns absolute filename of save file.

        :param str filename: Filename to storage file at.
        :param typing.Any data: Data to store to cache file. Can be any object.
        :return: Full filename to cached file.
        :rtype: str
        """

        self.check_cache_dir()
        path: str = self.build_cache_path(filename=filename)

        with open(path, "wb") as output_file:
            pickle.dump(data, output_file)

        return path


    def load(self, filename: str) -> str:
        """
        Load string from file.

        :param str filename: Filename of file to load from cache folder.
        :return: File content string.
        :rtype: str
        """

        path: str = self.build_cache_path(filename=filename)

        if not os.path.isfile(path):
            raise FileNotFoundError(f"Can not load file. File at path '{path}' does not exist.")

        with open(path, "r", encoding="utf-8") as f_obj:
            data = f_obj.read()
            f_obj.close()

        return data


    def load_dump(self, filename: str) -> Any:
        """
        Load binary dump from file.

        :param str filename: Filename of file to load from cache folder.
        :return: Object from file.
        :rtype: typing.Any
        """

        path: str = self.build_cache_path(filename=filename)

        if not os.path.isfile(path):
            raise FileNotFoundError(f"Can not load file. File at path '{path}' does not exist.")

        with open(path, "rb") as input_file:
            return pickle.load(input_file)



""" KEGG pathway models to parse object relational """
# pylint: disable=invalid-name,redefined-builtin,too-many-lines


# from warnings import warn
from datetime import datetime
from xml.etree import ElementTree
from xml.etree.ElementTree import Element
from typing import List, Union, Optional

from .const import (
    REACTION_TYPE,
    RELATION_TYPES,
    RELATION_SUBTYPES,
    ENTRY_TYPE,
    GRAPHIC_TYPE,
)

from .utils import (
    get_attribute,
    get_numeric_attribute,
    parse_xml,
    is_valid_pathway_number,
    is_valid_hex_color,
    is_valid_pathway_name,
    is_valid_pathway_org,
)




class Subtype:
    """
    Subtype model class.
    """

    def __init__(
        self,
        name: str,
        value: str,
    ) -> None:
        """
        Init Subtype model instance.

        :param str name: Name of subtype. Must match list of valid subtypes.
        :param str value: Value of subtype.
        """

        # check for valid subtype names in RELATION_SUBTYPES
        if name not in RELATION_SUBTYPES:
            raise ValueError(f"Name of relation subtype '{name}' is not in list of valid subtypes.")

        self.name: str = name
        self.value: str = value



    @staticmethod
    def parse(item: Element) -> "Subtype":
        """
        Parse Subtype XML element.

        :param xml.etree.ElementTree.Element item: XML element.
        :return: Parsed Subtype instance.
        :rtype: Subtype
        """

        # check correct type
        assert item.tag == "subtype"

        # Generate and return subtype instance
        return Subtype(
            name=get_attribute(element=item, key="name"),
            value=get_attribute(element=item, key="value")
        )


    def to_xml(self) -> Element:
        """
        Generate XML string from Subtype element.

        :return: XML string.
        :rtype: xml.etree.ElementTree.Element
        """

        return Element(
            "subtype",
            attrib={
                "name": self.name,
                "value": self.value
            },
        )


    def __str__(self) -> str:
        """
        Generate string from Subtype instance.

        :return: String of Subtype instance.
        :rtype: str
        """
        return f"<Subtype name='{self.name}' value='{self.value}'>"






class Relation:
    """
    Relation model class.
    """

    def __init__(
        self,
        entry1: str,
        entry2: str,
        type: str,
    ) -> None:
        """
        Init relation model instance.

        :param str entry1: Source entry of relation.
        :param str entry2: Destination entry of relation.
        :param str type: Type of Relation. Must be contained in list of valid relation types.
        """

        # Check if type is in valid relation types
        if type not in RELATION_TYPES:
            raise ValueError(f"Relation type '{type}' not in list of valid types.")

        self.entry1: str = entry1
        self.entry2: str = entry2
        self.type: str = type
        self.subtypes: List[Subtype] = []


    @staticmethod
    def parse(item: Element) -> "Relation":
        """
        Parse XML element into Relation instance.

        :param xml.etree.ElementTree.Element item: XML element to parse.
        :return: Relation instance.
        :rtype: Relation
        """

        # Check tag of relation is correct
        assert item.tag == "relation"


        # Create relation instance from attributes
        relation: Relation = Relation(
            entry1=get_numeric_attribute(element=item, key="entry1"),
            entry2=get_numeric_attribute(element=item, key="entry2"),
            type=get_attribute(element=item, key="type"),
        )


        # Parse Child items of xml element by iterating of child elements
        for child in item:
            relation.subtypes.append(Subtype.parse(item=child))


        return relation


    def to_xml(self) -> Element:
        """
        Generate XML string from Relation element.

        :return: XML string.
        :rtype: xml.etree.ElementTree.Element
        """

        return Element(
            "relation",
            attrib={
                "entry1": self.entry1,
                "entry2": self.entry2,
                "type": self.type,
            },
        )

    def __str__(self) -> str:
        """
        Generate string from relation instance.

        :return: String of Relation instance.
        :rtype: str
        """
        return f"<Relation {self.entry1}->{self.entry2} type='{self.type}'>"




class Component:
    """
    Component model.
    """

    def __init__(self, id: str) -> None:
        """
        Init Component model.

        :param str id: Id of component.
        """

        # id can't be empty
        if id == "":
            raise ValueError("Component id can't be empty.")


        # TODO: Check pattern of component id
        # TODO: component id should reference an existing entry !

        self.id: str = id


    @staticmethod
    def parse(item: Element) -> "Component":
        """
        Parsing ElementTree into Component.

        :param xml.etree.ElementTree.Element item: XML element to parse.
        :return: Parsed Component instance.
        :rtype: Component
        """

        # Check for correct xml tag
        assert item.tag == "component"

        # Create component instance from id attribute
        return Component(id=get_attribute(element=item, key="id"))


    def to_xml(self) -> Element:
        """
        Generate XML string from Component element.

        :return: XML string.
        :rtype: xml.etree.ElementTree.Element
        """

        return Element(
            "component",
            attrib={
                "id": self.id,
            },
        )


    def __str__(self) -> str:
        """
        Build string of component instance.

        :return: String of Component instance.
        :rtype: str
        """
        return f"<Component id='{self.id}'>"



class Graphics:
    """
    Graphics information for rendering.
    """

    def __init__(
        self,
        x: Optional[str] = None,
        y: Optional[str] = None,
        width: Optional[str] = None,
        height: Optional[str] = None,
        coords: Optional[str] = None,
        name: Optional[str] = None,
        type: Optional[str] = None,
        fgcolor: Optional[str] = None,
        bgcolor: Optional[str] = None,
    ) -> None:
        """
        Init Graphics model instance.

        :param typing.Optional[str] x:
        :param typing.Optional[str] y:
        :param typing.Optional[str] width:
        :param typing.Optional[str] height:
        :param typing.Optional[str] coords:
        :param typing.Optional[str] name:
        :param typing.Optional[str] type:
        :param typing.Optional[str] fgcolor:
        :param typing.Optional[str] bgcolor:
        """

        # All parameter are implied (optional)
        self.x: Optional[str] = x
        self.y: Optional[str] = y
        self.width: Optional[str] = width
        self.height: Optional[str] = height
        self.coords: Optional[str] = coords
        self.name: Optional[str] = name
        self.type: Optional[str] = type
        self.fgcolor: Optional[str] = fgcolor
        self.bgcolor: Optional[str] = bgcolor


        # Validate parameter if not None

        # Check type is in list of valid types
        if self.type is not None and self.type not in GRAPHIC_TYPE:
            raise ValueError(f"Type '{self.type}' is not a valid graphics type.")

        # Check foreground and background color are hex format
        if self.fgcolor is not None and is_valid_hex_color(value=self.fgcolor) is False:
            raise ValueError("Fgcolor is not a valid hex color.")

        if self.bgcolor is not None and is_valid_hex_color(value=self.bgcolor) is False:
            raise ValueError("Bgcolor is not a valid hex color.")


        # Check x,y,width and height are numeric (only if not None)
        if self.x is not None and str.isdigit(self.x) is False:
            raise ValueError("Value x is not a number.")

        if self.y is not None and str.isdigit(self.y) is False:
            raise ValueError("Value y is not a number.")

        if self.width is not None and str.isdigit(self.width) is False:
            raise ValueError("Value width is not a number.")

        if self.height is not None and str.isdigit(self.height) is False:
            raise ValueError("Value height is not a number.")


    @staticmethod
    def parse(item: Element) -> "Graphics":
        """
        Parse XML element into Graphics instance.

        :param xml.etree.ElementTree.Element item: XML element to parse.
        :return: Parsed Graphics instance.
        :rtype: Graphics
        """

        # Check xml tag
        assert item.tag == "graphics"

        # Parse attributes from XML element
        return Graphics(
            x = item.attrib.get("x"),
            y = item.attrib.get("y"),
            width = item.attrib.get("width"),
            height = item.attrib.get("height"),
            name = item.attrib.get("name"),
            type = item.attrib.get("type"),
            fgcolor = item.attrib.get("fgcolor"),
            bgcolor = item.attrib.get("bgcolor"),
            coords = item.attrib.get("coords"),
        )


    def to_xml(self) -> Element:
        """
        Generate XML string from Graphics element.

        :return: XML string.
        :rtype: xml.etree.ElementTree.Element
        """

        graphics_element: Element = Element(
            "graphics",
        )

        if self.x is not None:
            graphics_element.attrib["x"] = self.x

        if self.y is not None:
            graphics_element.attrib["y"] = self.y

        if self.width is not None:
            graphics_element.attrib["width"] = self.width

        if self.height is not None:
            graphics_element.attrib["height"] = self.height

        if self.name is not None:
            graphics_element.attrib["name"] = self.name

        if self.type is not None:
            graphics_element.attrib["type"] = self.type

        if self.fgcolor is not None:
            graphics_element.attrib["fgcolor"] = self.fgcolor

        if self.bgcolor is not None:
            graphics_element.attrib["bgcolor"] = self.bgcolor

        if self.coords is not None:
            graphics_element.attrib["coords"] = self.coords

        return graphics_element


    def __str__(self) -> str:
        """
        Return Graphics instance summary string.

        :return: String of Graphics instance.
        :rtype: str
        """
        return f"<Graphics name='{self.name}'>"



class Entry:
    """
    Entry model class.
    """

    def __init__(
        self,
        id: str,
        name: str,
        type: str,
        link: Optional[str] = None,
        reaction: Optional[str] = None,
    ) -> None:
        """
        Init entry model instance.

        :param str id: Id of Entry.
        :param str name: Name of Entry.
        :param str type: Type of Entry. Must be contained in list of valid entry types.
        :param typing.Optional[str] link: Link to KEGG database with reference to entry.
        :param typing.Optional[str] reaction: Reaction TODO: specify. Is str format correct?
        """

        # required
        self.id: str = id
        self.name: str = name
        self.type: str = type

        # TODO: validate entry id and name


        if self.type not in ENTRY_TYPE:
            raise ValueError(f"Type '{self.type}' is not in list of valid entry types.")


        # optional (implied)
        self.link: Optional[str] = link
        self.reaction: Optional[str] = reaction

        # TODO: validate reaction if not None


        # Implied child instances
        self.graphics: Optional[Graphics] = None
        self.components: List[Component] = []


    @property
    def has_multiple_names(self) -> bool:
        """
        Checks if entry has multiple names that are space seperated.

        :return: Retruns True if entry has multiple names.
        :rtype: bool
        """

        return len(self.name.split(" ")) > 1



    @staticmethod
    def parse(item: Element) -> "Entry":
        """
        Parsing XML element into Entry instance.

        :param xml.etree.ElementTree.Element item: XML element to parse.
        :return: Parsed Entry instance.
        :rtype: Entry
        """

        assert item.tag == "entry"

        # Generate entry instance from required attributes
        entry: Entry = Entry(
            id=get_numeric_attribute(element=item, key="id"),
            name=get_attribute(element=item, key="name"),
            type=get_attribute(element=item, key="type"),
            link=item.attrib.get("link"),
            reaction=item.attrib.get("reaction"),
        )

        # Iterate over children and parse graphics, components, ...
        for child in item:
            if child.tag == "graphics":
                entry.graphics = Graphics.parse(child)
            elif child.tag == "component":
                entry.components.append(Component.parse(child))

        return entry


    def to_xml(self) -> Element:
        """
        Generate XML string from Entry element.

        :return: XML string.
        :rtype: xml.etree.ElementTree.Element
        """

        entry_element: Element = Element(
            "entry",
            attrib={
                "id": self.id,
                "name": self.name,
                "type": self.type
            }
        )

        # Add optional attributes to xml element
        if self.link is not None:
            entry_element.attrib["link"] = self.link

        if self.reaction is not None:
            entry_element.attrib["reaction"] = self.reaction

        # generate xml element from children elements

        # Add graphics element to entry element
        if self.graphics is not None:
            entry_element.append(self.graphics.to_xml())


        # Iterate over components and add to entry element
        for component in self.components:
            entry_element.append(component.to_xml())


        return entry_element




    def __str__(self) -> str:
        """
        Build Entry summary string.

        :return: String of Entry instance.
        :rtype: str
        """
        return f"<Entry id='{self.id}' name='{self.name}' type='{self.type}'>"



    def get_gene_id(self) -> List[str]:
        """
        Parse variable 'name' of Entry into KEGG id.

        :return: List of KEGG identifier.
        :rtype: typing.List[str]
        """

        # TODO: validate return valid !!
        # r"^([a-z]){3}([0-9]){5}$"

        # return self.name.split(":")[1]
        return [value.split(":")[1] for value in self.name.split(" ")]




class Alt:
    """
    Alt model.
    """

    def __init__(self, name: str) -> None:
        """
        Init Alt instance.

        :param str name: Alt element name.
        """

        self.name: str = name


    @staticmethod
    def parse(item: Element) -> "Alt":
        """
        Parse Alt instance from XML element.

        :param xml.etree.ElementTree.Element item: XML element to parse.
        :rtype: Alt
        :return: Parsed Alt element.
        """
        assert item.tag == "alt"

        return Alt(name=get_attribute(element=item, key="name"))


    def to_xml(self) -> Element:
        """
        Generate XML string from Alt element.

        :return: XML string.
        :rtype: xml.etree.ElementTree.Element
        """

        return Element(
            "alt",
            attrib={
                "name": self.name,
            },
        )


    def __str__(self) -> str:
        """
        Build string from Alt instance.

        :return: String of Alt instance.
        :rtype: str
        """

        return f"<Alt name='{self.name}'>"


class Product:
    """
    Reaction Product model.
    """

    def __init__(
        self,
        id: str,
        name: str,
        alt: Optional[Alt] = None
        ) -> None:
        """
        Init Product instance.

        :param str id: Identifier of Product in pathway.
        :param str name: KEGG identifier of compound.
        :param Alt alt: Alternative name of element.
        """

        # TODO: verify correct format

        self.id: str = id
        self.name: str = name
        self.alt: Optional[Alt] = alt

    @staticmethod
    def parse(item: Element) -> "Product":
        """
        Parse XML element instance to Product model instance.

        :param xml.etree.ElementTree.Element item: XML element to parse.
        :return: Parsed Product model.
        :rtype: Product
        """

        assert item.tag == "product"

        parsed_product: Product = Product(
            id=get_attribute(element=item, key="id"),
            name=get_attribute(element=item, key="name"),
        )

        # Parse child alt elements
        for child in item:
            if child.tag == "alt":

                # TODO warn if overwrite
                # TODO: or raise exception (only 1 alt should be presnet)
                # if parsed_product.alt is not None:
                #     warn(message="'Alt'")

                parsed_product.alt = Alt.parse(item=child)

        return parsed_product


    def to_xml(self) -> Element:
        """
        Generate XML string from Product element.

        :return: XML string.
        :rtype: xml.etree.ElementTree.Element
        """

        product_element: Element = Element(
            "product",
            attrib={
                "id": self.id,
                "name": self.name,
            },
        )

        # Add optional alt element if exist
        if self.alt is not None:
            product_element.append(self.alt.to_xml())

        return product_element


    def __str__(self) -> str:
        """
        Build string from Product instance.

        :return: String of Product instance.
        :rtype: str
        """
        return f"<Product id='{self.id}' name='{self.name}'>"


class Substrate:
    """
    reaction Substrate model
    """
    def __init__(
        self,
        id: str,
        name: str,
        alt: Optional[Alt] = None,
        ) -> None:
        """
        Init Substrate instance.

        :param str id: Identifier of Substrate in pathway.
        :param str name: KEGG identifier of compound.
        :param Alt alt: Alternative name of element.
        """

        # TODO: verify correct format

        self.id: str = id
        self.name: str = name
        self.alt: Optional[Alt] = alt

    @staticmethod
    def parse(item: Element) -> "Substrate":
        """
        Parse XML element instance to Substrate model instance.

        :param xml.etree.ElementTree.Element item: XML element to parse.
        :return: Parsed Substrate model.
        :rtype: Substrate
        """

        assert item.tag == "substrate"

        parsed_substrate: Substrate = Substrate(
            id=get_attribute(element=item, key="id"),
            name=get_attribute(element=item, key="name"),
        )

        # Parse child alt elements
        for child in item:
            if child.tag == "alt":

                # TODO warn if overwrite
                # TODO: or raise exception (only 1 alt should be presnet)
                # if parsed_substrate.alt is not None:
                #     warn(message="'Alt'")

                parsed_substrate.alt = Alt.parse(item=child)

        return parsed_substrate


    def to_xml(self) -> Element:
        """
        Generate XML string from Substrate element.

        :return: XML string.
        :rtype: xml.etree.ElementTree.Element
        """

        substrate_element: Element = Element(
            "substrate",
            attrib={
                "id": self.id,
                "name": self.name,
            },
        )

        # Add optional alt element if exist
        if self.alt is not None:
            substrate_element.append(self.alt.to_xml())

        return substrate_element


    def __str__(self) -> str:
        """
        Build string from Substrate instance.

        :return: String of Substrate instance.
        :rtype: str
        """
        return f"<Substrate id='{self.id}' name='{self.name}'>"



class Reaction:
    """
    Reaction model.
    """

    def __init__(
        self,
        id: str,
        name: str,
        type: str,
        ) -> None:
        """
        Init Reaction model instance.

        :param str id: Identifier of reaction.
        :param str name: KEGG identifer of reaction.
        :param str type: Type of reaction. Must be contained in list of valid reaction types.
        """

        if type not in REACTION_TYPE:
            raise ValueError("Type of reaction is not in list of valid reaction types.")

        # TODO: check valid reaction id
        self.id = id
        self.name = name
        self.type = type

        # Child elements of reaction
        self.products: List[Product] = []
        self.substrates: List[Substrate] = []


    @staticmethod
    def parse(item: Element) -> "Reaction":
        """
        Parse XML element instance to Reaction model instance.

        :param xml.etree.ElementTree.Element item: XML element to parse.
        :return: Parsed Reaction model.
        :rtype: Reaction
        """

        assert item.tag == "reaction"

        # Parse reaction instance from xml attributes
        parsed_reaction: Reaction = Reaction(
            id=get_attribute(element=item, key="id"),
            name=get_attribute(element=item, key="name"),
            type=get_attribute(element=item, key="type"),
        )

        # Parse product and substrate from child elements
        for child in item:
            if child.tag == "product":
                parsed_reaction.products.append(Product.parse(child))
            elif child.tag == "substrate":
                parsed_reaction.substrates.append(Substrate.parse(child))

        return parsed_reaction


    def to_xml(self) -> Element:
        """
        Generate XML string from Reaction element.

        :return: XML string.
        :rtype: xml.etree.ElementTree.Element
        """

        reaction_element: Element = Element(
            "reaction",
            attrib={
                "id": self.id,
                "name": self.name,
                "type": self.type,
            },
        )

        # Add substrate and product to reaction element

        for substrate in self.substrates:
            reaction_element.append(substrate.to_xml())

        for product in self.products:
            reaction_element.append(product.to_xml())


        return reaction_element


    def __str__(self) -> str:
        """
        Build string of reaction instance.

        :return: String of Reaction instance.
        :rtype: str
        """

        return f"<Reaction id='{self.id}' name='{self.name}'>"




class Pathway:
    """
    KEGG Pathway object.
    The KEGG pathway object stores graphics information and related objects.
    """
    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        name: str,
        org: str,
        number: str,
        title: Optional[str] = None,
        image: Optional[str] = None,
        link: Optional[str] = None,
        ) -> None:
        """
        Init KEGG Pathway model.

        :param str name: Name of pathway, which is the full KEGG identifier.
        :param str org: Organism code.
        :param str number: Number of pathway.
        :param typing.Optional[str] title: Title of pathway.
        :param typing.Optional[str] image: Image for pathway provided by KEGG database.
        :param typing.Optional[str] link: Link to pathway in KEGG database.
        """

        # required parameter of pathway element
        self.name: str = name
        self.org: str = org
        self.number: str = number

        # Check all required parameter formats
        if not is_valid_pathway_name(value=name):
            raise ValueError(f"Pathway name '{name}' is not a valid value.")

        if not is_valid_pathway_org(value=org):
            raise ValueError(f"Pathway org '{org}' is not a valid value.")

        if not is_valid_pathway_number(value=number):
            raise ValueError(f"Pathway number '{number}' is not a valid value.")


        # Check match of number, org and name
        if self.name != f"path:{self.org}{self.number}":
            raise ValueError("Mismatch of arguments name, org and number.")


        # implied
        self.title: Optional[str] = title
        self.image: Optional[str] = image
        self.link: Optional[str] = link

        # children
        self.relations: List[Relation] = []
        self.entries: List[Entry] = []
        self.reactions: List[Reaction] = []




    @staticmethod
    def parse(data: Union[Element, str]) -> "Pathway":
        """
        Parsing XML string or element in Pathway instance.

        :param typing.Union[xml.etree.ElementTree.Element, str] data: String or XML element to parse.
        :return: Parsed Pathway instance.
        :rtype: Pathway
        """

        # Generate correct format from string or XML element object
        item: Element = parse_xml(xml_object_or_string=data)


        # Init pathway instance with all required attributes
        pathway: Pathway = Pathway(
            name=get_attribute(element=item, key="name"),
            org=get_attribute(element=item, key="org"),
            number=get_attribute(element=item, key="number"),
            title=item.attrib.get("title"),
            image=item.attrib.get("image"),
            link=item.attrib.get("link"),
        )


        # Parse child items of pathway
        for child in item:
            if child.tag == "entry":
                pathway.entries.append(Entry.parse(child))
            elif child.tag == "relation":
                pathway.relations.append(Relation.parse(child))
            elif child.tag == "reaction":
                pathway.reactions.append(Reaction.parse(child))


        return pathway


    def to_xml(self) -> Element:
        """
        Generate XML element from Pathway instance and its children.

        :return: XML element in KGML format.
        :rtype: xml.etree.ElementTree.Element
        """

        pathway_element: Element = Element(
            "pathway",
            attrib={
                "name": self.name,
                "org": self.org,
                "number": self.number,
            }
        )

        # Adding optional attributes to pathway root element
        if self.title is not None:
            pathway_element.attrib["title"] = self.title

        if self.link is not None:
            pathway_element.attrib["link"] = self.link

        if self.image is not None:
            pathway_element.attrib["image"] = self.image


        # add children entries, relations and reactions

        for entry in self.entries:
            pathway_element.append(entry.to_xml())

        for relation in self.relations:
            pathway_element.append(relation.to_xml())

        for reaction in self.reactions:
            pathway_element.append(reaction.to_xml())


        return pathway_element


    def to_xml_string(self) -> str:
        """
        Generate XML string from pathway instance.

        :return: XML string in KGML format.
        :rtype: str
        """

        # Generate xml header string
        # docstring is not supported by build-in xml builer
        xml_timestamp: str = datetime.now().astimezone().strftime("%b %d, %Y %H:%M:%S (GMT%z)")
        xml_header: str = "<?xml version=\"1.0\"?>\n" \
            "<!DOCTYPE pathway SYSTEM \"http://www.kegg.jp/kegg/xml/KGML_v0.7.2_.dtd\">\n" \
            f"<!-- Creation date: {xml_timestamp} -->\n"


        pathway_element: Element = self.to_xml()

        xml_content: str = ElementTree.tostring(pathway_element).decode("utf-8")

        return xml_header + xml_content


    def get_entry_by_id(self, entry_id: str) -> Optional[Entry]:
        """
        Get pathway Entry object by id.

        :param str entry_id: Id of Entry.
        :return: Returns Entry instance if id is found in Pathway. Otherwise returns None.
        :rtype: typing.Optional[Entry]
        """

        for item in self.entries:
            if item.id == entry_id:
                return item
        return None



    def get_genes(self) -> List[str]:
        """
        List all genes from pathway.

        :return: List of entry ids with type gene.
        :rtype: typing.List[str]
        """

        result: List[str] = []

        # Iterate of entries and get all gene type
        # Keep list of genes unique
        for entry in self.entries:
            if entry.type == "gene":

                # Get name of entry (KEGG identifier)
                # if " " in entry.name:
                #     # Check if name contains a space, which indicates list of multiple identifier
                #     splitted_entries: List[str] = entry.name.split(" ")
                #     for single_entry in splitted_entries:
                #         if single_entry not in result:
                #             result.append(single_entry)


                # elif entry.name not in result:
                #     result.append(entry.name)

                for gene_id in entry.get_gene_id():
                    if gene_id not in result:
                        result.append(gene_id)

        return result


    def __str__(self) -> str:
        """
        Build string summary for KEGG pathway.

        :return: String of Pathway instance.
        :rtype: str
        """
        return f"<Pathway path:{self.org}{self.number} title='{self.title}'>"


    def merge(self) -> "Pathway":
        """
        Merge identical entries in Pathway together and generate new pathway instance.

        :return: Merged Pathway instance.
        :rtype: Pathway
        """

        merged_pathway: Pathway = Pathway(
            name=self.name,
            org=self.org,
            number=self.number,
            title=self.title,
        )

        # TODO find duplicate entries in pathway

        return merged_pathway



""" Basic utils for HTTP requests, parsing and rendering """

import re
import csv
from io import StringIO

from typing import Dict, List, Optional, Union
from xml.etree import ElementTree
from xml.etree.ElementTree import Element


# XML parsing helper functions


def get_attribute(element: Element, key: str) -> str:
    """
    Get attribute from XML Element object. Raises KeyError is Attribute is not found or not valid.

    :param xml.etree.ElementTree.Element element: XML element to get attribute from.
    :param str key: Name of attribute.
    :return: Value of attribute.
    :rtype: str
    :raises ValueError: Error if attribute does not exist or is wrong type.
    """

    value: Optional[str] = element.attrib.get(key)

    # Check if value is not none and is string
    if value is None or isinstance(value, str) is False:
        raise ValueError(f"Value of attribute '{key}' is not a string.")

    return value



def get_numeric_attribute(element: Element, key: str) -> str:
    """
    Get attribute from XML Element object. Raises KeyError is Attribute is not found or not valid.

    :param Element element: XML element to get attribute from.
    :param str key: Name of attribute.
    :return: Value of attribute. ValueError is raised if value is not a numeric string.
    :rtype: str
    :raises ValueError: Error is attribute is not a digit (numberic string), does not exist or is wrong type.
    """

    value: str = get_attribute(element=element, key=key)

    # Check if string is numeric
    if str.isnumeric(value) is False:
        raise ValueError(f"Value of attribute '{key}' is not numeric.")

    return value



def parse_xml(xml_object_or_string: Union[str, Element]) -> Element:
    """
    Returns XML Element object from string or XML Element.

    :param typing.Union[str, xml.etree.ElementTree.Element] xml_object_or_string: Input parameter to check.
    :return: XML element instance.
    :rtype: xml.etree.ElementTree.Element
    """
    if isinstance(xml_object_or_string, str):
        return ElementTree.fromstring(xml_object_or_string)
    return xml_object_or_string




def parse_tsv(data: str) -> list:
    """
    Parse .tsv file from string

    :param str data: Tsv string to parse into list.
    :return: List of items.
    :rtype: list
    """
    return list(csv.reader(StringIO(data), delimiter="\t"))



def parse_tsv_to_dict(
    data: str,
    col_keys: int = 0,
    col_values: int = 1,
    ) -> Dict[str, str]:
    """
    Parse .tsv file from string and build dict from first two columns. Other columns are ignored.

    :param str data: Tsv string to parse.
    :param int col_keys: Number of colum to parse as dict keys (0-index).
    :param int col_values: Number of colum to parse as dict values (0-index).
    :return: Dict of two tsv columns.
    :rtype: typing.Dict[str, str]
    """
    list_data: list = parse_tsv(data=data)

    result: Dict[str, str] = {}

    for row in list_data:
        if len(row) >= 2 and row[col_keys] != "":
            result[row[col_keys]] = row[col_values]

        # TODO: handle else cases or ignore silent ?

    return result



class ColorGradient:
    """
    Create color gradient.
    """

    def __init__(
        self,
        start: tuple,
        stop: tuple,
        steps: int = 100
    ) -> None:
        """
        Init ColorGradient instance.

        :param tuple start: Color tuple
        :param tuple stop: Color tuple
        :param int steps: Number of steps.
        """
        self.start: tuple = start
        self.stop: tuple = stop
        self.steps: int = steps


    @staticmethod
    def to_css(color: tuple) -> str:
        """
        Convert color tuple to CSS rgb color string.

        :param tuple color: RGB color tuple containing 3 integers
        :return: Color as CSS string (e.g. "rgb(0, 0, 0)").
        :rtype: str
        """
        # if len(color) != 3 or not all([isinstance(value, int) for value in color]):
        #     raise ValueError("Color must be a tuple of 3 integers.")

        return f"rgb({color[0]},{color[1]},{color[2]})".lower()


    @staticmethod
    def to_hex(color: tuple) -> str:
        """
        Convert color tuple to hex color string.

        :param tuple color: RGB color tuple containing 3 integers.
        :return: Hexadecimal color string (e.g. "#000000").
        :rtype: str
        """

        # TODO: check for int type.
        # TODO: check for 0-255 range

        return f"#{color[0]:02x}{color[1]:02x}{color[2]:02x}".lower()


    def get_list(self) -> List[str]:
        """
        Get gradient color as list.

        :return: Returns list of hexadecimal color strings with a gradient.
        :rtype: typing.List[str]
        """

        step_list = [index / float(self.steps) for index in range(self.steps)]
        result = [ColorGradient._intermediate(self.stop, self.start, step) for step in step_list]
        result.append(self.stop)


        return [
            ColorGradient.to_hex(code) for code in result
        ]


    @staticmethod
    def _intermediate(a_var: tuple, b_var: tuple, ratio: float) -> tuple:
        def _array_multiply(array, c_var):
            return [element * c_var for element in array]

        a_component = _array_multiply(a_var, ratio)
        b_component = _array_multiply(b_var, 1 - ratio)
        values: List[float] = list(map(sum, zip(a_component, b_component)))

        return tuple((int(item) for item in values))





def is_valid_pathway_org(value: str) -> bool:
    """
    Check if organism identifier is valid.

    :param str value: String value to check.
    :return: Returns True if value is a valid organism code.
    :rtype: bool
    """

    # Organism must be 3 letter code
    # Identifier can also be KO or Enzyme identifer
    # TODO: validate with KEGG organism list
    return re.match(pattern=r"^(ko|ec|[a-z]{3})$", string=value) is not None


def is_valid_pathway_number(value: str) -> bool:
    """
    Check if pathway number has correct 5 digit format.

    :param str value: String value to check.
    :return: Returns True if value has the correct format of pathway number.
    :rtype: bool
    """

    # KEGG pathway number must be a 5 digit number
    return re.match(pattern=r"^([0-9]{5})$", string=value) is not None


def is_valid_pathway_name(value: str) -> bool:
    """
    Check if combined pathway identifer is valid. String must match "path:<org><number>".

    :param str value: String value to check.
    :return: Returns True if value matches format of pathway name.
    :rtype: bool
    """

    return re.match(pattern=r"^path:(ko|ec|[a-z]{3})([0-9]{5})$", string=value) is not None


def is_valid_hex_color(value: str) -> bool:
    """
    Check if string is a valid hex color.

    :param str value: String value to check.
    :return: Returns True if value is valid hexadecimal color string.
    :rtype: bool
    """
    return re.match(pattern=r"^\#([a-fA-F0-9]{6})$", string=value) is not None


def is_valid_gene_name(value: str) -> bool:
    """
    Check if gene identifer is valid. String must match "<org>:<number>".

    :param str value: String value to check.
    :return: Returns True if value matches format of gene name.
    :rtype: bool
    """

    # TODO Support ko|ec entries ???
    return re.match(pattern=r"^([a-z]{3}):([0-9]{5})$", string=value) is not None



# TODO: is a custom warning class needed
# class MissingDataWarning(Warning):
#     """
#     Warning to alert user of missing data.
#     """

#     def __init__(self, message: str) -> None:
#         """
#         Init custom missing data warning instance.

#         :param str message: Message of missing data warning.
#         """

#         # Call super init method
#         super().__init__()

#         self.message: str = message


#     def __str__(self) -> str:
#         """
#         Missing data warning instance to string.

#         :return: String of message.
#         :rtype: str
#         """

#         return repr(self.message)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['keggtools']

package_data = \
{'': ['*']}

install_requires = \
['pydot >= 1.4.2', 'requests >= 2.27.1', 'scipy >= 1.5.4']

extras_require = \
{'doc': ['Sphinx', 'sphinx-rtd-theme'],
 'test': ['pylint',
          'pytest',
          'coverage',
          'mypy',
          'types-requests',
          'responses',
          'pandas']}

setup(name='keggtools',
      version='1.0.0',
      description='Enrichment analysis and visualisation toolkit for KEGG pathways',
      author=None,
      author_email='harryhaller001 <harryhaller001@gmail.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )
